/**
 *  @file BasicEightMap.h
 *
 *  @author kapela  @date 2010-12-06
 */

#ifndef KAPELA_BASICBasicEightMap_H_
#define KAPELA_BASICBasicEightMap_H_

//#include "capd/matrixAlgorithms/floatMatrixAlgorithms.hpp"
#include "capd/normalForms/linearSubstitution.hpp"
#include "nbody.h"
#include "capd/poincare/CoordinateSection.h"
#include "capd/poincare/PoincareMap.hpp"

//#include "capd/dynsys/TaylorHOE.hpp"

template <typename PoincareMapT>
class BasicEightMap {
public:
  typedef PoincareMapT PoincareMapType;
  typedef typename PoincareMapType::Solver DynamicalSystemType;
  typedef typename DynamicalSystemType::VectorFieldType MapType;  
  typedef typename MapType::FunctionType FunctionType;

  typedef typename DynamicalSystemType::ScalarType ScalarType;
  typedef typename DynamicalSystemType::VectorType VectorType;
  typedef typename DynamicalSystemType::MatrixType MatrixType;
  typedef typename DynamicalSystemType::StepControlType StepControlType;
  typedef typename MapType::JetType CnCoeffType;

  typedef typename capd::poincare::CoordinateSection<MatrixType> SectionType;
  
protected:
  MapType vf;
  DynamicalSystemType dynsys;
  SectionType section;
  PoincareMapType poincareMap;
  FunctionType H;
  MapType expansion;
  ScalarType M;
  ScalarType h;
  VectorType Z;
  int order;
  int maxDerivativeOrder;
public:

  BasicEightMap(
      const ScalarType & M,                      ///< angular momentum
      const ScalarType & energyLevel,            ///< energy level
      int                taylorOrder,
      const StepControlType & stepControl = StepControlType(),
      int maxDerivativeOrder = 1
    ) :
        vf(whittaker3BodyEquation2D(), maxDerivativeOrder),
        dynsys(vf, taylorOrder, stepControl),
  //      section("var:q1,q2,q3,p1,p2,p3;fun:q3;"),
        section(6, 2),
        poincareMap(dynsys, section, capd::poincare::MinusPlus),
        H(energy(M)),
        expansion(getExpansion(M, energyLevel, maxDerivativeOrder)),
        M(M),
        h(energyLevel),
        order(taylorOrder),
        maxDerivativeOrder(maxDerivativeOrder)
  {
    vf.setParameter("M",M);
  }

//  void setParametersForDerivativesComputations(
//      int taylorOrder,
//      const ScalarType  & timeStep
//  ) {
//    orderDer = taylorOrder;
//    stepDer = timeStep;
//  }

  int dimension() const {
    return 4;
  }

  DynamicalSystemType & getDynSys() { return dynsys; }

  void setAngularMomentum(const ScalarType & new_M){
    M = new_M;
    vf.setParameter("M",M);
    H.setParameter("M",M);
    expansion.setParameter("M",M);
  }

  void setEnergyLevel(const ScalarType & new_h){
    h = new_h;
    expansion.setParameter("h",h);
  }
  
  /// Returns function that for given configuration computes energy 
  FunctionType energy(const ScalarType & M){
    FunctionType H("var:q1,q2,q3,p1,p2,p3;par:M;"
                   "fun:p1^2+p2^2+p3^2+p1*p2-p3/q1*(q2*p3-q3*p2-M)+1/q1^2*(q2*p3-q3*p2-M)^2-1/q1-(q2^2+q3^2)^(-0.5)-((q1-q2)^2+q3^2)^(-0.5);");
    H.setParameter("M",M);
    return H;
  }
  
  MapType getExpansion(const ScalarType & M, const ScalarType & h, int rank){
    MapType expansion(
        "var:q1,q2,q3,p1,p2,p3;par:M,h;"
        "fun:q1,q2,0,p1,p2,"
        "(-(M/q1-2*M*q2/(q1^2))+sqrt((M/q1-2*M*q2/(q1)^2)^2-4*(1-q2/q1+(q2/q1)^2)*((p1)^2+(p2)^2+p1*p2+(M/q1)^2-1/q1-1/sqrt(q2^2)-1/sqrt((q1-q2)^2)-h)))/(2*(1-q2/q1+(q2/q1)^2));",
        rank
    );
    expansion.setParameter("M",M);
    expansion.setParameter("h",h);
    return expansion;
  }
  
// a =  p3^2-p3/q1*(q2*p3)+q2^2*p3^2/q1^2
// b =  -1/q1*(-M)+1/q1^2*(2*q2*(-M)
// c=  p1^2+p2^2+p1*p2+1/q1^2*(-q3*p2-M)^2-1/q1-(q2^2+q3^2)^(-0.5)-((q1-q2)^2+q3^2)^(-0.5)

  VectorType expandIC(const VectorType & x){
    VectorType x0(6);
    x0[0] = x[0];            // q1
    x0[1] = x[1];            // q2
//    x0[2] = ScalarType(0.0);   // q3 (it will be computed by expansion)
    x0[3] = x[2];            // p1
    x0[4] = x[3];            // p2
//    x0[5] = z1;              // p3 (it will be computed by expansion)
    x0 = expansion(x0);
    
    return x0;
  }
  VectorType reduce(const VectorType & x){
      VectorType x0(4);
      x0[0] = x[0];            // q1
      x0[1] = x[1];            // q2
      x0[2] = x[3];            // p1
      x0[3] = x[4];            // p2
      return x0;
  }
  MatrixType reduce(const MatrixType & D){
    MatrixType A(4, 4);
    int p[] = {0, 1, 3, 4};
    for(int i=0; i<4; ++i){
      for(int j=0; j<4; ++j){
        A[i][j] = D[p[i]][p[j]];
      }
    }
    return A;
  }

  /**
   * We reduce map: R^6 -> R^6  to  R^4 -> R^4
   *
   * @param coeff    series of map with variables (q1,q2,q3,p1,p2,p3)
   * @return         series in variables  (q1,q2,p1,p2)
   */
  CnCoeffType reduce(const CnCoeffType & coeff){
   // We permute coordinates so that coordinates that should remain are on the first places
   VectorType zero(6);
   MapType permutation("var:q1,q2,q3,p1,p2,p3;fun:q1,q2,p1,p2,q3,p3;");
   MatrixType P = permutation[zero];
//   assertContainsZero(P*Transpose(P)-IMatrix::Identity(6));

   CnCoeffType result = capd::normalForms::linearSubstitution(P, coeff, Transpose(P));
   capd::sbug << "\n coefficient before reduction after permutation of variables :" << result.toString(0,6,0,0,3);

   // After permutation multipointers for result and reduced are the same.
   CnCoeffType reduced(4, coeff.degree());
   for(int r = 0; r <= coeff.degree(); ++r) {
     typename CnCoeffType::Multipointer mp = reduced.first(r);
     do {
       for(int i = 0; i < reduced.dimension(); ++i) {
         reduced(i, mp) = result(i, mp);
       }
     } while(reduced.hasNext(mp));
   }
   capd::sbug << "\n coefficient after reduction :" << reduced.toString();

   return reduced;
  }

  VectorType compute(const VectorType & x, MatrixType & derivative){

    capd::sbug << "\n --- computing value and derivative ---\n x : " << x ;

    VectorType x0 = expandIC(x);
    MatrixType derExpansion = expansion[x0];

    MatrixType der(6,6);

    VectorType result = poincareMap(x0, der);
    capd::sbug << "\n final result : " << result
               << "\n reduced      : " << reduce(result);

    // testFirstIntegralsPreservation(x0, result);

    der = poincareMap.computeDP(result, der);
    derivative = reduce(der*derExpansion);

    return reduce(result);
  }

  VectorType operator()(
      const VectorType & x,
      MatrixType &       derivative
  ){
    return compute(x, derivative);
  }

  MatrixType operator[](const VectorType & x){
    MatrixType derivative(x.dimension(), x.dimension());
    compute(x, derivative);
    return derivative;
  }

  VectorType operator()(const VectorType & x){

    VectorType x0=expandIC(x);
    VectorType result = poincareMap(x0);
    return reduce(result);
  }

};

#endif
