/**
 *  @file EightMap.h
 *
 *  @author kapela  @date 2010-12-06
 */

#ifndef KAPELA_EIGHTMAP_H_
#define KAPELA_EIGHTMAP_H_

#include "BasicEightMap.h"

template <typename PoincareMapT>
class EightMap : public BasicEightMap<PoincareMapT> {
public:
  typedef BasicEightMap<PoincareMapT> BaseType;
  typedef PoincareMapT PoincareMapType;
  typedef typename PoincareMapType::Solver DynamicalSystemType;
  typedef typename DynamicalSystemType::VectorFieldType MapType;
  typedef typename MapType::FunctionType FunctionType;

  typedef typename DynamicalSystemType::ScalarType ScalarType;
  typedef typename DynamicalSystemType::VectorType VectorType;
  typedef typename DynamicalSystemType::MatrixType MatrixType;
  typedef typename DynamicalSystemType::StepControlType StepControlType;
  typedef typename MapType::JetType CnCoeffType;

  typedef typename capd::poincare::CoordinateSection<MatrixType> SectionType;
  
protected:
  using BaseType::vf;
  using BaseType::dynsys;
  using BaseType::section;
  using BaseType::poincareMap;
  using BaseType::H;
  using BaseType::expansion;
  using BaseType::M;
  using BaseType::h;
  using BaseType::order;
  int orderDer;
public:

  EightMap(
      const ScalarType & M,                      ///< angular momentum
      const ScalarType & energyLevel,            ///< energy level
      int                taylorOrder,
      const StepControlType & stepControl = StepControlType(),
      int maxDerivativeOrder = 1
  ) : BaseType(M, energyLevel, taylorOrder, stepControl, maxDerivativeOrder), 
      orderDer(taylorOrder) {
  }
  EightMap(const ODEParameters<VectorType> & ode,
        int order = 20,               ///< taylor order for solver
        int maxDerivativeOrder = 1,
        const StepControlType & stepControl = StepControlType() 
  ) : BaseType(ode.param, ode.h, order, stepControl, maxDerivativeOrder),
      orderDer(order){ 
  }
  VectorType computeRigorous(const VectorType & x, MatrixType & derivative, int numberOfIterates = 1){

    capd::sbug << "\n --- computing value and derivative ---\n x : " << x 
               << "\n diam x : " << capd::vectalg::maxWidth(x);

    VectorType x0 = this->expandIC(x);
    MatrixType derExpansion = this->expansion[x0];

    C1Set::C0BaseSet c0part(x0);
    C1Set::C1BaseSet c1part(derExpansion);
    C1Set set (c0part, c1part);
    MatrixType der(6,6);

    //    capd::sbug << "\n x0 : " << x0
    //           << "\n c0part : " << c0part.toString()
    //           << "\n c1part : " << c1part.toString()
    //           << "\n set    : " << set.show();

    for(int i=0; i<numberOfIterates-1; i++){
      IVector res = poincareMap(set);
    }

    IVector result = poincareMap(set, der);

    capd::sbug << "\n final result : " << result 
               << "\n  diam : " << maxWidth(result)
               << "\n reduced : " << this->reduce(result) ;

    //testFirstIntegralsPreservation(x0, result);

    der = poincareMap.computeDP(result, der);
    //    std::cout << "\n przed redukcja : " << der;
    derivative = this->reduce(der);
    //    std::cout << "\n po redukcji : " << derivative;
    return this->reduce(result);
  }
  VectorType operator()(const VectorType & x, int numberOfIterates = 1){
    dynsys.setOrder(order);
    VectorType x0= this->expandIC(x);
    C0Set set(x0);

    for(int i=0; i<numberOfIterates-1; i++){
      IVector res = poincareMap(set);
    }
    IVector result = poincareMap(set);

    capd::sbug << "\n final result : " << result << "\n  diam : " << diam(result)
                           << "\n zredukowany : " << this->reduce(result) ;

    return this->reduce(result);
  }
  MatrixType operator[](const VectorType & x){
    dynsys.setOrder(orderDer);
    MatrixType der(x.dimension(), x.dimension());
    computeRigorous(x, der);
    return der;
  }


  VectorType operator()(
      const VectorType & x,
      MatrixType &       derivative,
      int numberOfIterates = 1
  ){
    dynsys.setOrder(order);
    return computeRigorous(x, derivative);
  }

  /**
   *
   */
  VectorType computeCn(
      const VectorType & x,     ///< initial conditions
      int                rank,  ///< maximal order of derivatives
      CnCoeffType &      deriv  ///< @param[out] derivatives of Poincare Map up to order 'rank'
  ){

    capd::sbug << "\n --- computing value and derivative ---\n x : " << x << "\n diam x : " << diam(x);
    VectorType x0 = this->expandIC(x);
    int dim = x0.dimension();
    CnCoeffType derExpansion(dim, rank);
    expansion(x0, derExpansion);

    CnSet set (derExpansion);


    CnCoeffType der(dim, rank);
    IVector result = poincareMap(set, der, 3);

    capd::sbug << "\n final result : " << result << "\n  diam : " << diam(result)
                         << "\n reduced : " << this->reduce(result) ;

    capd::sbug << "\n derivative of the flow \n" << der;
    //testFirstIntegralsPreservation(x0, result);

    der = poincareMap.computeDP(der);

    capd::sbug << "\n derivative of the Poincare map \n" << der;

    deriv = this->reduce(der);
  
    return this->reduce(result);
  }

  /**
   *
   */
  VectorType operator()(
      const VectorType & x,     ///< initial conditions
      CnCoeffType &      deriv,  ///< @param[out] derivatives of Poincare Map up to order 'rank'
      int numberOfIterates = 1
  ){

    capd::sbug << "\n --- computing value and derivative ---\n x : " << x << "\n diam x : " << diam(x);

    int rank = deriv.degree();
    VectorType x0 = this->expandIC(x);
    int dim = x0.dimension();
    CnCoeffType derExpansion(dim, rank);
    expansion(x0, derExpansion);

    CnSet set (derExpansion);

    CnCoeffType der(dim, rank);
    IVector Px0 = poincareMap(set, der, numberOfIterates);
    IVector Px = this->reduce(Px0);

    capd::sbug << "\n final result : " << Px0 << "\n  diam : " << diam(Px0)
                               << "\n reduced : " << Px ;

    capd::sbug << "\n derivative of the flow \n" << der;
    //testFirstIntegralsPreservation(x0, result);

    der = poincareMap.computeDP(der);

    capd::sbug << "\n derivative of the Poincare map \n" << der;

    //    std::cout << "\n przed redukcja : " << der;
    deriv = this->reduce(der);
    //    std::cout << "\n po redukcji : " << derivative;

    return Px;
  }



};

#endif /* CHOREOMAP_H_ */
