/**
 *  @file ChoreoMap.h
 *
 *  @author kapela  @date 2010-12-06
 */

#ifndef PARALLELMAP_H_
#define PARALLELMAP_H_
#include "capd/vectalg/SubVector.hpp"
#include "capd/matrixAlgorithms/floatMatrixAlgorithms.hpp"
#include "capd/matrixAlgorithms/capd2alglib.h"


template <typename T, bool isInterval = capd::TypeTraits<typename T::ScalarType>::isInterval>
class MaxWidth{
  public: static typename capd::TypeTraits<typename T::ScalarType>::Real value(const T & v){
    return 0;
  }
};
template <typename T>
class MaxWidth<T, true>{
  public: static typename capd::TypeTraits<typename T::ScalarType>::Real value(const T & v){
    return capd::vectalg::maxWidth(v);
  }
};
template <typename IMatrix, typename IVector>
IVector computeNewtonOperator2(const IVector & x0, const IVector & fx0, const IMatrix & dfX, int period){
  int dim_full = x0.dimension();
  int dim = dim_full/period;
  IMatrix dgX(dim_full, dim_full);
  IVector gx0(dim_full);
  for(int i=0; i<period; ++i){
    for (int j = i*dim; j < (i+1)*dim; ++j) {
      for (int k = 0; k < dim; ++k) {
        dgX[j][i*dim+k] = dfX[j][k];
      }
      dgX[j][(j+dim)%dim_full] = typename IVector::ScalarType(-1.0);
      gx0[j] = fx0[j] - x0[(j+dim)%dim_full];
    }
  }
  std::cout  << "\n fx0" << fx0 << "\n dfX " << dfX <<  "\n gx0 " << gx0 << "\n dgX" << dgX ;
  IVector N = x0 - capd::matrixAlgorithms::gauss(dgX, gx0);
  std::cout << "\n---- computeNewtonOperator2 ---- \n x0 " << x0 << "\n N " << N << "\n--------------\n";
  return N;
}

template <typename PoincareMapT>
class ParallelMap {
public:
  typedef PoincareMapT PoincareMapType;
  typedef typename PoincareMapType::DynamicalSystemType DynamicalSystemType;
  typedef typename DynamicalSystemType::VectorFieldType MapType;
  typedef typename MapType::FunctionType FunctionType;

  typedef typename DynamicalSystemType::ScalarType ScalarType;
  typedef typename DynamicalSystemType::VectorType VectorType;
  typedef typename DynamicalSystemType::MatrixType MatrixType;
  typedef typename MapType::JetType CnCoeffType;

protected:

  PoincareMapType & poincareMap;
  int N;         ///< number of sections
  int dim;       ///< dimension of phase space
  int nDim;      ///< dimension in parallel shooting = N * dim

public:

  ParallelMap(
      PoincareMapType & poincareMap,       ///< poincare map
      int N=3                              ///< number of iterates of Poincare Map
  ) :
    poincareMap(poincareMap),
    N(N),
    dim(poincareMap.dimension()),
    nDim(N * dim) {
  }

  int dimension() const {
    return nDim;
  }
  int parallelDim() const {
    return nDim;
  }
  int originalDim() const {
    return dim;
  }
  VectorType operator()(const VectorType & x, MatrixType & derivative){
    if(x.dimension() != nDim)
      throw std::runtime_error("ParallelMap: wrong input vector dimension!");
    return compute(x, derivative);
  }

  MatrixType operator[](const VectorType & x){
    if(x.dimension() != nDim)
      throw std::runtime_error("ParallelMap: wrong input vector dimension!");
    MatrixType der(nDim , dim);
    compute(x, der);
    return der;
  }

  VectorType operator()(const VectorType & x){

    if(x.dimension() != nDim)
      throw std::runtime_error("ParallelMap: wrong input vector dimension!");

    VectorType Px(dimension());

    for (int i = 0; i < N; ++i) {
      VectorType xi(dim);
      std::copy(x.begin()+i*dim, x.begin()+(i+1)*dim, xi.begin());

      VectorType Pxi =  poincareMap(xi);

      std::copy(Pxi.begin(), Pxi.end(), Px.begin()+i*dim);
    }
    return Px;
  }
  /**
   *  computes image and derivatives (up to given order)
   *  of composition of all poincare maps
   */
  VectorType computeCn(
      const VectorType & x,   ///< initial condition
      CnCoeffType & dPx       ///< @param[in,out]   on return it contains derivatives up to order given by it method 'rank()'
  ){

     int rank = dPx.degree();
     VectorType Pxi(x.dimension());
 
    // poincareMap.turnOnStepControl();
     capd::vectalg::EuclLNorm<VectorType, MatrixType> norm;

     for (int i = 0; i < N; ++i) {
       VectorType xi(dim);
       std::copy(x.begin()+i*dim, x.begin()+(i+1)*dim, xi.begin());

       
       CnCoeffType dPxi(dim, rank);
       Pxi = poincareMap(xi, dPxi);

      if(i==0)
        dPx = dPxi;
      else{
        CnCoeffType dPx1 = dPx;
        substitutionPowerSeries(dPxi, dPx1, dPx, false);
      }
//      std::cout << "\n\nresult \n" << dPx.toString(0, dPx.dimension(), 0, 0, dPx.degree(), capd::TypeTraits<Interval>::numberOfDigits()+1 );
     }

     dPx() = Pxi;
//     std::cout << "\n\nresult \n" << dPx.toString(0, dPx.dimension(), 0, 0, dPx.degree(), capd::TypeTraits<Interval>::numberOfDigits()+1 );

     return Pxi;
   }

  /// finds approximated initial condition for parallel shooting
  ///  by computing iterates of Poincare Map of x
  VectorType findIC(
      const VectorType & x    ///<  approximated initial condition on the first section
  ){
    capd::scon << "\n---------------\nFinding initial condition for parallel shoooting " << std::endl;

    VectorType result(nDim);
    std::copy(x.begin(), x.end(), result.begin());
    Vector x0(x);

    for(int i=1; i<N; i++){
      capd::sbug << "\n computing poincare map " << i ;

      x0 = poincareMap(x0);
      std::copy(x0.begin(), x0.end(), result.begin()+dim*i);

      capd::sbug << "\n    " << i << " - result : " << x0 ;
    }

    capd::scon << "\ninitial condition for parallel shooting :\n  "  << result << std::endl;

    return result;
  }

  /// Computes Poincare Map for each iteration
  VectorType compute(const VectorType & x, MatrixType & dPx){

      VectorType Px(nDim);
      dPx = MatrixType(nDim, dim);

      for (int i = 0; i < N; ++i) {
        VectorType xi(dim);
        std::copy(x.begin()+i*dim, x.begin()+(i+1)*dim, xi.begin());

        MatrixType dPxi(dim, dim);
        capd::sbug << "\n computing poincare map " << i ;

        VectorType Pxi =  poincareMap(xi, dPxi);

        // We expect that poincareMap already corrected derivatives
        // dPxi = poincareMap.computeDP(Pxi, dPxi);

        capd::sbug << "\n   Pxi : " << Pxi << "\n-----";

        std::copy(Pxi.begin(), Pxi.end(), Px.begin()+i*dim);

        capd::vectalg::copyToSubmatrix(dPxi, dPx, dim*i, dim*(i+1), 0, dim);
      }

      capd::sbug << "\n poincare computed\n";
      return Px;
    }

  VectorType computeNewtonOperator(const VectorType & x0, const VectorType & Px0, const MatrixType & dPX, ScalarType & error){
    using std::max;
    using std::abs;
    MatrixType DF(nDim, nDim);
    VectorType F(nDim);
    for(int i=0; i<nDim; ++i){
      F[i] = Px0[i] - x0[(i+dim)%nDim];
      DF(i+1,(i+dim)%nDim+1) = ScalarType(-1.0);
    }
    int row=1;
    for(int offset=0; offset<nDim; offset += dim){
      for(int r=0; r<dim; ++r, ++row){
        for(int c=1; c<=dim; ++c){
          DF(row,offset+c) = dPX(row, c);
        }
      }
    }
    capd::vectalg::MaxNorm<VectorType, MatrixType> maxNorm;
    error = maxNorm(F);
    
    capd::sbug << "\n max error = " << error;
    capd::sbug << "\n F " << F<< "\n dF " << DF;
    VectorType z0 = capd::matrixAlgorithms::gauss(DF, F);
    VectorType N = x0 - z0;
    
    ScalarType error2;
    VectorType N2 = computeNewtonOperator_inSmallDim(x0,Px0, dPX, error2);
    capd::sbug << "\n\nN1 = " << N 
              << "\n\nN2 = " << N2 
              << "\n\nmax width N1 = " << MaxWidth<VectorType>::value(N) 
              << "\n\nmax width N2 = " << MaxWidth<VectorType>::value(N2)
            << "\n\n N1 - N2 = " << (N - N2) << std::endl;
    return N;
  }
  
  VectorType computeNewtonOperator_inSmallDim(const VectorType & x0, const VectorType & Px0, const MatrixType & dPX, ScalarType & error){
    using std::max;
    using std::abs;
    int dimN = dimension();
    MatrixType D = MatrixType::Identity(dim);
    VectorType r(dim);
    std::vector<VectorType> f(N);
    std::vector<MatrixType> Di(N);
    error = 0;
    for(int i = N-1; i>=0; --i){
      capd::sbug << "\n---------------------";
      f[i].resize(dim);
      for(int j=0; j<dim; ++j){
        f[i][j] = Px0[dim*i+j]- x0[((i+1)*dim+j)%dimN];;
        capd::sbug << "\n " << Px0[dim*i+j] << "  -  " <<  x0[((i+1)*dim+j)%dimN] << "  =  " <<  f[i][j];

        error = capd::max(error, capd::abs(f[i][j]));
      }
      capd::sbug << "\n---------------------";

      Di[i].resize(dim,dim);
      capd::vectalg::copyFromSubmatrix(dPX,Di[i],dim*i,dim*(i+1),0,dim);
//      std::cout << "  \n D" << i << " = " << Di[i];
      r += D * f[i];
      D = D * Di[i];
    }
    capd::sbug << "\n max error = " << error;
    capd::sbug << "\n r " << r << "\n D " << D;
    VectorType z0 = capd::matrixAlgorithms::gauss(D - MatrixType::Identity(dim), r);
    VectorType z(dimN);
    std::copy(z0.begin(), z0.end(), z.begin());
    for(int i=0; i<N-1; ++i){
      z0 = Di[i]*z0 - f[i];
      std::copy(z0.begin(), z0.end(), z.begin()+dim*(i+1));
    }
    VectorType N = x0 - z;

    return N;
  }

  /// refines initial condition (in full space N*dim)
  VectorType refineIC(
      const VectorType & x,       ///<  approx. i.c
      int                maxN,    ///<  maximal number of iterations
      ScalarType         maxError///<  error tolerance
  ){
    if(maxN <=0)
      return x;
    MatrixType dPx(nDim,dim);
    return refineIC(x, maxN, maxError, dPx);
  }

  /// refines initial condition (in full space N*dim)
  /// @return it returns i.c x=(x0, x1,... xn) so that \f$ |P(x_i) - x_{i+1}] < maxError \f$  if maxN iteration was not achieved
  VectorType refineIC(
      const VectorType & x,       ///<  approx. i.c
      int                maxN,    ///<  maximal number of iterations
      ScalarType         maxError,///<  error tolerance
      MatrixType &       dPx      ///<  on return dPx is (N*dim) x dim matrix that contains derivatives of poincare maps
  ){
    
    capd::scon << "\n-----------------\nRefining initial condition\n  error tolerance : " << maxError  << "   max number of iterations : " << maxN << std::endl;
    VectorType x0 = x;
    int iter=0;
    ScalarType error;
    VectorType Px;
    do{
      Px = compute(x0, dPx);
        capd::sbug << "\n computing newton operator"<< "\n x0 " << x0 << "\n Px " << Px << "\n dPx" << dPx;
      VectorType x1 = computeNewtonOperator(x0, Px, dPx, error);
      x0 = x1;
      iter++;
        capd::sbug << " iteration nr. " << iter << "  error : " << error << std::endl;
    }while((error>maxError)&&(iter<maxN));

    capd::scon << "\n refined i.c. : " << x0 << "\n error : " << error << std::endl;
    return x0;
  }
  /**
   *
   *  Computes arguments of eigenvalues of real 4x4 matrix
   */
  void computeEigenvalues4D(const MatrixType & dPX,  ScalarType & alpha1, ScalarType & alpha2){
    using std::max;
    MatrixType D = MatrixType::Identity(dim);
    MatrixType Di(dim,dim);

    for(int i = N-1; i>=0; --i){
      capd::vectalg::copyFromSubmatrix(dPX,Di,dim*i,dim*(i+1),0,dim);
      D = D * Di;
    }

    capd::vectalg::Matrix<double,0,0> A(dim,dim);
    for(int i=0; i<dim; ++i)
      for(int j=0; j<dim; ++j)
        A[i][j] = D[i][j];

    capd::vectalg::Vector<double,0> im(dim), re(dim);
    capd::alglib::computeEigenvalues(A, re, im);
    double m_2PI =  6.28318530717958647692;
    //std::cout << "\n\n eigenvalues : " << capd::alglib::eigenvaluesToString(re, im);
    //std::cout << "\n\n sigma 1 " <<
    alpha1 = std::atan2(im[0], re[0]) / m_2PI;
    //std::cout << "\n\n sigma 1 " <<
    alpha2 = std::atan2(im[2], re[2]) / m_2PI;
     if(alpha1>alpha2)
       std::swap(alpha1, alpha2);
  }
};

#endif /* CHOREOMAP_H_ */
