/**
 *  @file Parameters.h
 *
 *  @date   Mar 10, 2011
 *  @author kapela
 */

#ifndef PARAMETERS_H_
#define PARAMETERS_H_
#include "capd/auxil/ConfigFileReader.h"

template <typename IVectorT>
struct ODEParameters{
  typedef IVectorT IVector;
  typedef typename IVector::ScalarType Interval;
  typedef typename Interval::BoundType Float;
  typedef typename IVector::template rebind<Float>::other Vector;
  
  Interval h;
  Interval param;
  int period;
  int precision;
  Vector x0;           // initial conditions 
  Vector parallel_x0;  // initial conditions for parallel shooting
  IVector X;           // rigorous estimates of initial conditions (in parallel shooting)
  
  ODEParameters()
    : h(0.5,0.5), param(-0.9,-0.9), 
       period(3), precision(52),
      x0(2), parallel_x0(2), X(2) {
  }
  bool readParameters(
      capd::auxil::ConfigFileReader & ini,
      const std::string & section = ""
  ) {
    int precision = 100;
    ini.getValue(section+"precision", precision);
   // setPrecision(precision);
    bool isOK = true;
    isOK = ini.getValue(section+"h", h) && isOK;
    isOK = (ini.getValue(section+"param", param) || ini.getValue(section+"M", param) || ini.getValue("c", param)) && isOK;
    isOK = (ini.getValue(section+"period", period) || ini.getValue(section+"period", period)) && isOK;
    isOK = ini.getValue(section+"x0", x0) && isOK;
    ini.getValue(section+"parallel_x0", parallel_x0);
    ini.getValue(section+"X",X);
    if(!isOK) capd::sbug << "\nODEParameters: one of parameters is missing! \n\n "; 
    return isOK;
  }
  friend std::ostream & operator << (std::ostream & str, const ODEParameters & o){
    str     << "h = " << o.h
            << "\nparam = " << o.param
            << "\nperiod = " << o.period
            << "\nprecision = " << o.precision
            << "\n# Approximated initial conditions"
            << "\nx0 = " << o.x0;
    if(o.parallel_x0 != Vector(o.parallel_x0.dimension()))
        str << "\n# Approximated initial condition for parallel shooting"
            << "\nparallel_x0 = " << o.parallel_x0;
    if(o.X != IVector(o.X.dimension()))
        str << "\n# Rigorous enclosures for initial conditions"
            << "\nX = " << o.X;
    str << std::endl;
    return str;
  }
};

template <typename T>
struct Parameters{
public:
  int order;
 // T step;
  int precision;
  std::string outputFile;
  
  Parameters(): 
    order(20), 
   // step(1./1024.), 
    precision(52), 
    outputFile("") {
  }
  
  void readParameters(
        capd::auxil::ConfigFileReader & ini,                   ///< parsed ini file
        const std::string & section = ""   ///< section name
  ){
      ini.getValue(section+"order", order);
     // ini.getValue(section+"step", step);
      ini.getValue(section+"precision", precision);
      ini.getValue(section+"output", outputFile);
  }
  friend std::ostream & operator << (std::ostream & str, const Parameters & o){
    str << "order = " << o.order
       // << "\nstep = " << o.step
        << "\nprecision = " << o.precision
        << std::endl;
    if(o.outputFile != "")
      str << "output = " << o.outputFile << std::endl;
    return str;
  }
};

template <typename T>
struct RefineParameters : public Parameters<T> {
  int maxNumberOfIterations;
  T errorTolerance;
  
  RefineParameters() 
    : Parameters<T>(), 
      maxNumberOfIterations(5), 
      errorTolerance(1.e-13) {
  }

  void readParameters(
      capd::auxil::ConfigFileReader & ini,                          ///< parsed ini file
      const std::string & section = "refine."   ///< section name
  ){
    Parameters<T>::readParameters(ini, section);
    ini.getValue(section+"maxNumberOfIterations", maxNumberOfIterations);
    ini.getValue(section+"errorTolerance", errorTolerance);
  }
  friend std::ostream & operator << (std::ostream & str, const RefineParameters & o){
    str << "[refine]\n"
        << static_cast<const Parameters<T> & >(o)
        << "maxNumberOfIterations = " << o.maxNumberOfIterations
        << "\nerrorTolerance = " << o.errorTolerance
        << std::endl;
    return str;
  }
};

template <typename T>
struct ExistenceProofParameters : public Parameters<T>{
  int orderDer;
  T stepDer;
  T boxSize;
  T errorTolerance;
  int numberOfImprovements;

  ExistenceProofParameters() 
        : Parameters<T>(),
          orderDer(20),
          stepDer(1./1024.),
          boxSize(1e-10),
          errorTolerance(1.e-11),
          numberOfImprovements(5){
  }
  void readParameters(
      capd::auxil::ConfigFileReader & ini,                          ///< parsed ini file
      const std::string & section = "existence."   ///< section name
  ){
    Parameters<T>::readParameters(ini, section);
    ini.getValue(section+"orderDer", orderDer);
    ini.getValue(section+"stepDer", stepDer);
    ini.getValue(section+"boxSize", boxSize);
    ini.getValue(section+"box size", boxSize);
    ini.getValue(section+"errorTolerance", errorTolerance);
    ini.getValue(section+"maxNumberOfIterations", numberOfImprovements);
    ini.getValue(section+"numberOfImprovements", numberOfImprovements);
    
  }
  friend std::ostream & operator << (std::ostream & str, const ExistenceProofParameters & o){
    str << "[existence]\n"
        << static_cast<const Parameters<T> & >(o)
        << "orderDer = " << o.orderDer
        << "\nstepDer = " << o.stepDer
        << "\nboxSize = " << o.boxSize
        << "\nmaxNumberOfIterations = " << o.numberOfImprovements
        << "\nerrorTolerance = " << o.errorTolerance
        << std::endl;
    return str;
  }
};

template <typename T>
struct ComputeCnParameters : public Parameters<T> {
  bool computeParallel;

  ComputeCnParameters()
          : Parameters<T>(),
          computeParallel(true){
  }
  
  void readParameters(
      capd::auxil::ConfigFileReader & ini,                          ///< parsed ini file
      const std::string & section = "computeCn."   ///< section name
  ){
    Parameters<T>::readParameters(ini, section);
    ini.getValue(section+"computeParallel", computeParallel);
  }
  friend std::ostream & operator << (std::ostream & str, const ComputeCnParameters & o){
    str << "[computeCn]\n"
        << static_cast<const Parameters<T> & >(o)
        << "computeParallel = " << o.computeParallel
        << std::endl;
    return str;
  }
};

// ?? order, step ??

template <typename T>
struct KAMProofParameters : public Parameters<T> {

  void readParameters(
      capd::auxil::ConfigFileReader & ini,                          ///< parsed ini file
      const std::string & section = "KAMProof."   ///< section name
  ){
    Parameters<T>::readParameters(ini, section);
  }
  friend std::ostream & operator << (std::ostream & str, const KAMProofParameters & o){
    str << "[KAMProof]\n"
        << static_cast<const Parameters<T> & >(o)
        << std::endl;
    return str;
  }
};

template <typename IVectorT>
struct ParametersSet {
  typedef IVectorT IVector;
  typedef typename IVector::ScalarType Interval;
  typedef typename Interval::BoundType Float;
  ODEParameters<IVector> ode;
  RefineParameters<Float> refine;
  ExistenceProofParameters<Float> existenceProof;
  ComputeCnParameters<Float> computeCn;
  KAMProofParameters<Float> KAMProof;

  bool readParameters(
      capd::auxil::ConfigFileReader & ini                          ///< parsed ini file
  ){
    //ini.setStopTag("raw_data");
    if(!ode.readParameters(ini)) return false;
    refine.readParameters(ini);
    existenceProof.readParameters(ini);
    computeCn.readParameters(ini);
    KAMProof.readParameters(ini);
    return true;
  }
  friend std::ostream & operator << (std::ostream & str, const ParametersSet & o){
    str <<
"#================================================ \n"
"# ODE parameters and initial conditions            \n"
"#================================================\n"
<< o.ode << 
"\n#================================================\n"
"# parameters used to refine approximated i.c.    \n"
"#================================================\n"
<< o.refine <<
"\n#================================================\n"
"# existence proof parameters                     \n"
"#================================================\n"
<< o.existenceProof <<
"\n#================================================\n"
"# parameters for Cn computations\n"
"#================================================\n"
<< o.computeCn <<
"\n#================================================\n"
"# parameters for KAM proof\n"
"#================================================\n"
<< o.KAMProof 
        << std::endl;
    return str;
  }
};

#endif /* PARAMETERS_H_ */
