/* 
 * File:   H4Proof.h
 * Author: kapela
 *
 * Created on 3 listopad 2014, 11:03
 */

#ifndef PROOF_H
#define	PROOF_H

#include "typedefs.h"
#include "ioUtils.h"
#include "improveIC.h"
#include "existenceProof.h"
#include "Parameters.h"
#include "CnCoeffUtils.h"
#include "normalFormComputation.h"
#include "capd/basicalg/TexWriter.h"

template <typename MapType, typename ODEParameters, typename RefineParameters>
Vector improveIC(ODEParameters & ode, const RefineParameters & refine){
  setPrecision(refine.precision);
  
  MapType dpm(leftBound(ode.param), leftBound(ode.h), refine.order);
  
  return refineICParallel(dpm, ode.period, ode.x0, refine.maxNumberOfIterations, refine.errorTolerance);
}

template <typename MapType, typename ODEParameters, typename ExistenceProofParameters>
IVector proveExistence( ODEParameters & ode, const ExistenceProofParameters & params) { 

  
  
  setPrecision(params.precision);
  double error = power(10, -capd::TypeTraits<Interval>::numberOfDigits()-3);

  
 // H4Map<PoincareMap> h4(c, h, existenceParams.order, existenceParams.step, capd::dynsys::ILastTermsStepControl(1, error, 1. / 1048576.));
  MapType h4(ode, params.order);
  h4.getDynSys().setAbsoluteTolerance(error);
  h4.getDynSys().setRelativeTolerance(error);
  
  capd::scon << "\n----------------\nExistence proof"
             //<< params
             << "\n  precision      : " << params.precision
             << "\n  order          : " << params.order 
             << "\n  approx.i.c. x0 : " << ode.parallel_x0 
             << "\n  size of set X  :" << params.boxSize << std::endl;

  IVector result = makeExistenceProof(
                          h4, ode.period, ode.parallel_x0, params.boxSize, 
                          params.errorTolerance, params.numberOfImprovements
                   );

  if(params.outputFile != "")
    saveData(params.outputFile.c_str(), ode.h, ode.param, ode.period, result, "Rigorous initial condition");
  
  return result;
}

template <typename MapType, typename ODEParameters, typename ComputeCnParameters>
ICnCoeff computeCn( ODEParameters & ode, const ComputeCnParameters & params, int orderOfDerivatives = 3) { 
//(IniParser & ini, const Interval & h, const Interval & c, IVector x0, int period ){
  
  capd::scon << "\n----------------\nComputing Taylor Expansion of the Poincare Map"
             << "\n  precision      : " << params.precision
             << "\n  order          : " << params.order 
             << "\n  i.c. X         : " << ode.X 
             << std::endl;
    
  setPrecision(params.precision);
  
  MapType map(ode, params.order, orderOfDerivatives);
  
  //  double error = power(10, -capd::TypeTraits<Interval>::numberOfDigits()-3);
  
  ICnCoeff dPx(map.dimension(), orderOfDerivatives);
  
  if(params.computeParallel){ 
    // Computes each iteration separately and then composes derivatives
    ParallelMap<MapType> pm(map, ode.period);
    pm.computeCn(ode.X, dPx);
  } else {
    // Computes derivatives in one long integration
    map(ode.X, dPx, ode.period);
  }
 
  capd::sbug << "\n computed jet of Poincare Map : \n" << dPx.toString()<< std::endl;
  // TODO: Compare those two computations (improve second by replacing C0 information before each period)
  
  if(params.outputFile != "")
     saveData(params.outputFile.c_str() , dPx, ode.h, ode.param, ode.period, ode.X, "Cn data");
  return dPx;
}


bool proofKAMstability2D(
    const ICnCoeff & cncoeff,   
    capd::TexWriter & output 
) {

  capd::sout << "\n Initial C^n jet computed on periodic orbit \n";

//  capd::map::print(std::cout, cncoeff) <<  diamToString(cncoeff, false) << std::endl
//      << "\n Taylor series : ";
 // printTaylorSeries(std::cout, cncoeff)<< std::endl;;

  CIVector eigenvalues = computeEigenvalues2D((IMatrix)cncoeff);

  output.precision(8) << "\n eigenvalues  \\[" << eigenvalues <<"\\]\n";
  capd::sout << "\n Rigorous estimates of eigenvalues : \n  " << eigenvalues << std::endl;

  CICnCoeff normalForm = computeNormalForm(cncoeff, eigenvalues);

  capd::sout << "\\\\\n Normal form :\n ";
//  capd::map::print(std::cout, normalForm) << std::endl;
  output.precision(3) << "\n Normal form (non zero coefficients of order 3)\n\n";
  writeNonZeroCoeff(output, normalForm, 3);
  writeNonZeroCoeff(capd::sout, normalForm, 3);
  output << "\\\\\n Torsion \\\\\n";

  return checkTwistCondition2D(normalForm, eigenvalues , output);
}

bool proofKAMstability4D(
    const ICnCoeff & cncoeff,
    capd::TexWriter & output
) {

  capd::sout << "\n Initial C^n jet computed on periodic orbit \n";

//  capd::map::print(std::cout, cncoeff) <<  diamToString(cncoeff, false) << std::endl
//      << "\n Taylor series : ";
 // printTaylorSeries(std::cout, cncoeff)<< std::endl;;

  CIVector eigenvalues = computeEigenvalues4D((IMatrix)cncoeff);

  output.precision(8) << "\n eigenvalues  \\[" << eigenvalues <<"\\]\n";
  capd::sout << "\n Rigorous estimates of eigenvalues : \n  " << eigenvalues << std::endl;

  CICnCoeff normalForm = computeNormalForm(cncoeff, eigenvalues);
  capd::sout << "\\\\\n Normal form (non zero coefficients of order 3) :\n ";
  //printTaylorSeries(std::cout, normalForm) << std::endl;
  output.precision(3) << "\n Normal form (non zero coefficients of order 3) \n\n";
  writeNonZeroCoeff(output, normalForm, 3);
  writeNonZeroCoeff(capd::sout, normalForm, 3, 10);
  output << "\\\\\n Torsion \\\\\n";

  return checkTwistConditionEight(normalForm, eigenvalues , output);
}
template <typename ODEParameters>
bool proofKAMstability(
    ODEParameters & ode,
    const ICnCoeff & cncoeff,
    const std::string & texFile
) {
  std::ofstream texStream(texFile);
  capd::TexWriter output(texStream);
  output.setBaseStyle(1).setFloatStyle(capd::TexWriter::FloatFix).setISymbol(capd::TexWriter::ImSymbolii);
  output.writeDocumentHeader("\\def\\ii{\\mbox{i\\,}} \n"
  "\\def\\iv#1{\\protect{\\makebox{\\boldmath $#1$}}} \n"
  "\\newcommand{\\x}{\\iv{x} }");
  output.precision(16);
  output << "\n energy level = $" << leftBound(ode.h) << "$"
         << "\\\\\n Initial conditions \\\\\n"         
          << "\n\n% \\x = $" <<  ode.X << "$\n";
//  output.setVectorStyle(capd::TexWriter::VerticalVector) 
  output.setVectorStyle(2) 
          << "\n\n \\[ \\x = " <<  ode.X << "\\]\n" ;
  output.precision(3) << "\n\n diam \\x= $" << maxWidth(ode.X) <<"$\n\n";
         
//  output.setVectorStyle(capd::TexWriter::HorizontalVector);
  output.setVectorStyle(0);
 
  bool result = false;
  switch(cncoeff.dimension()){
    case 2:
      result = proofKAMstability2D(cncoeff, output);
      break;
    case 4:
      result = proofKAMstability4D(cncoeff, output);
      break;
    default:
      throw std::runtime_error("proofKAMstability: method not implemented for given dimension ! ");
  }
  output.writeDocumentFooter();
  return result;
}


template <typename NonRigorousMap, typename RigorousMap, typename RigorousCnMap>
void makeCompleteProof(int argc, char **argv){
  std::string dataFile = "", resultFile = "", texFile = "output.tex";
  ProofStage startStage = ProofStage::Refine; 
  
  if(readCommandLineParametersWithStage(argc, argv, dataFile, resultFile, texFile, startStage ))
    return;

  //capd::auxil::ConfigFileReader ini(dataFile);
  capd::auxil::ConfigFileReader ini;
  ICnCoeff coeff;
  readConfigFile(ini, dataFile, startStage, &coeff);
  
  ParametersSet<IVector> params;
  params.readParameters(ini);

  switch(startStage){
    case ProofStage::Refine : {
      Vector dx0 = improveIC<NonRigorousMap> (params.ode, params.refine);
      // we store 
      params.ode.parallel_x0 = dx0;
      for(int i=0; i<params.ode.x0.dimension(); ++i)
        params.ode.x0[i] = dx0[i];
    }

    case ProofStage::ExistenceProof: {
      IVector  X = proveExistence<RigorousMap>(params.ode, params.existenceProof);
      params.ode.X = X;
    }

    case ProofStage::CnComputation : {
      coeff = computeCn<RigorousCnMap>(params.ode, params.computeCn);
    }
    
    case ProofStage::KAMProof : {
      proofKAMstability(params.ode, coeff, texFile);
    }
  }
  saveProofResults(resultFile, params, &coeff);  
}
#endif	/* H4PROOF_H */

