/**
 *  @file ioUtils.h
 *
 *  @author kapela  @date 2011-01-13
 */

#ifndef IOUTILS_H_
#define IOUTILS_H_
#include "capd/auxil/logger_deprecated.h"
#include "capd/auxil/arg.h"
#include <sstream>
#include <stdexcept>
#include <memory>
#include "capd/auxil/ConfigFileReader.h" 
#include "Parameters.h"
#include "typedefs.h"


/// sets default precision for Float and appropriate precision of std::cout
inline void setPrecision(int precision){
#ifndef _USE_DOUBLES_
  capd::MpFloat::setDefaultPrecision(precision);
#endif
  std::cout.precision(capd::TypeTraits<Interval>::numberOfDigits()+1);
//  std::cout << "\n -prec = " << std::cout.precision() << std::endl;
}
/// set appropriate precision to display full information stored in Float variables
inline void setPrecision(std::ostream & str){
  str.precision(capd::TypeTraits<Interval>::numberOfDigits()+1);
  str << std::fixed;
  capd::sbug << "\n prec = " << str.precision() << std::endl;
}

/// returns number of mantissa bits used by Float
inline int getPrecision(){
#ifndef _USE_DOUBLES_
  return capd::MpFloat::getDefaultPrecision();
#else
  return 52;
#endif
}

/**
 * Reads file and returns its content without comments.
 * Comments can start with #, ; or %
 *
 * @param name file name
 */
std::string  readAndSkipComments(const char * name);

/**
 * Load data from file.
 * File structure should be:
 * \code
 * h
 * M
 * N
 * x0
 * \endcode
 * Additional comments starting with '#', ';' or '%' are allowed.
 */
template <typename VectorType>
void loadData(
    const char * name,                    ///< file name
    typename VectorType::ScalarType & h,  ///< energy level
    typename VectorType::ScalarType & M,  ///< angular momentum
    int & N,                              ///< period (number of iterates of Poincare Map)
    VectorType & x0                       ///< initial condition
){
  std::istringstream str(readAndSkipComments(name));
  str >> h >> M >> N >> x0;
}

template <typename VectorType>
void saveData(
    std::ostream & file,                        ///< output stream
    const typename VectorType::ScalarType & h,  ///< energy level
    const typename VectorType::ScalarType & M,  ///< angular momentum or c
    int N,                                      ///< period (number of iterates of Poincare Map)
    const VectorType & x0,                      ///< initial condition
    const char * comment = ""                   ///< comment
){
  int precision = getPrecision();
  file << "# " << comment
       << "\n\n# h - energy level \nh = " << h
       << "\n\n# M (angular momentum) or c parameter \nparam = " << M
       << "\n\n# N - period (number of iterations of Poincare Map) \nperiod = " << N
       << "\n\n# x0 - initial condition \nx0 = " << x0
       << "\n\nprecision = " << precision
       << std::endl;
}



template <typename VectorType>
void saveData(
    const char * name,                          ///< file name
    const typename VectorType::ScalarType & h,  ///< energy level
    const typename VectorType::ScalarType & M,  ///< angular momentum or c
    int N,                                      ///< period (number of iterates of Poincare Map)
    const VectorType & x0,                      ///< initial condition
    const char * comment = ""                   ///< comment
){
  std::ofstream file(name);
  if(!file){
    std::cerr << "\n Error when openning file \"" << name << "\" for writting!" << std::endl;
    std::runtime_error("Error when openning file in saveData functions. ");
  }
  setPrecision(file);
  saveData(file, h, M, N, x0, comment);
}

template <typename VectorType, typename CnCoeffType>
void saveData(
    const char * name,                          ///< file name
    const CnCoeffType & coeff,                  ///< Cn coefficients
    const typename VectorType::ScalarType & h,  ///< energy level
    const typename VectorType::ScalarType & M,  ///< angular momentum or c
    int N,                                      ///< period (number of iterates of Poincare Map)
    const VectorType & x0,                      ///< initial condition
    const char * comment = ""                   ///< comment
){
  std::ofstream file(name);
  if(!file){
    std::cerr << "\n Error when openning file \"" << name << "\" for writting!" << std::endl;
    std::runtime_error("Error when openning file in saveData functions. ");
  }
  file << std::fixed;
  setPrecision(file);
  coeff.save(file);
  file << std::endl;
  saveData(file, h, M, N, x0, comment);
}



template <typename VectorType, typename ScalarType>
bool getODEParameters(
    capd::auxil::ConfigFileReader & ini,                      ///< parsed ini file
    ScalarType & h,                       ///< energy level
    ScalarType & param,                   ///< angular momentum (M) or prameter c
    int & N,                              ///< period (number of iterates of Poincare Map)
    VectorType & x0                       ///< initial condition
){
  int precision = 100;
  ini.getValue("precision", precision);
  setPrecision(precision);
  bool isOK = true;
  isOK = ini.getValue("h", h)  && isOK;
  isOK = (ini.getValue("param", param) || ini.getValue("M", param) || ini.getValue("c", param) ) && isOK;
  isOK = (ini.getValue("N", N) || ini.getValue("period",N)) && isOK;
  isOK = ini.getValue("x0", x0) && isOK;
  return isOK;
}

extern const char *default_helpinfo;

/**
 *  Parses command line arguments 
 */
int readCommandLineParameters (
        int argc, char ** argv,                 ///< comand line arguments
        std::string & dataFile,                 ///< input data file 
        std::string & resultTexFile,            ///< file where TeX output will be written
        const char * help_info= default_helpinfo   ///< help information displayed if '-h' argument is given
   );

/// Proof stages 
enum class ProofStage{ 
  Refine=1, ExistenceProof, CnComputation, KAMProof
};

/**
 *  Parses command line arguments 
 */
int readCommandLineParametersWithStage (
        int argc, char ** argv,                 ///< comand line arguments
        std::string & dataFile,                 ///< input data file 
        std::string & outputFile,               ///< output file  
        std::string & resultTexFile,            ///< file where TeX output will be written
        ProofStage & stage,                     ///< from which stage the computation should start
        const char * help_info = 0              ///< help information displayed if '-h' argument is given
   );


std::shared_ptr<std::ifstream> readConfigFile(
    capd::auxil::ConfigFileReader & ini, 
    const std::string & dataFile,        
    ProofStage stage = ProofStage::Refine,
    ICnCoeff * coeff = nullptr
);

template <typename ParametersType>
void saveProofResults(const std::string & resultFile, 
        const ParametersType & params,
        ICnCoeff * coeff = nullptr      
){
  if(resultFile != ""){
    capd::sout << "\n Writting output to file : " << resultFile << std::endl;
    std::ofstream output(resultFile.c_str());
    if(!output){
      capd::serr << "\n Error when creating file " << resultFile << std::endl;
      exit(-1);
    }
    setPrecision(output);
    output << params ;
    if(coeff != nullptr){
      output << "\n[raw_data]\n" 
            << *coeff << std::endl;
    }
    output.close();
  }
}
#endif /* IOUTILS_H_ */
